;DEVELOPED BY ROMUX.COM
#DEFINE	M_LCD_ID dummy

;***** INCLUDE FILES *****



;***** CONSTANT DECLARATION *****

	IFNDEF	LCDLINENUM	; use default value, if unspecified
		CONSTANT LCDLINENUM = 0x02	; by default, 2 lines
	ENDIF
	IFNDEF	LCDTYPE		; use default value, if unspecified
		CONSTANT LCDTYPE = 0x00		; standard HD44780 LCD
		;CONSTANT LCDTYPE = 0x01	; EADIP204-4 (w/ KS0073)
	ENDIF
	IFNDEF	LCDSPEED	; use default value, if unspecified
		;CONSTANT LCDSPEED = 0x00	; clk in [0..9] MHz
		CONSTANT LCDSPEED = 0x01	; clk in [9..20] MHz, default
	ENDIF
	IFNDEF	LCDWAIT		; use default value, if unspecified
		CONSTANT LCDWAIT = 0x05	  ; for Tosc <= 5 MHz
	ENDIF
	IFNDEF	LCDCLRWAIT	; use default value, if unspecified
		CONSTANT LCDCLRWAIT = 0x050	; wait after LCDCLR until LCD is ready again
	ENDIF
	IFNDEF	LCDBUSYWAIT	; use default value, if unspecified
		CONSTANT LCDBUSYWAIT = 0x50	; wait before LCD busy flag is available
	ENDIF

;***** HARDWARE DECLARATION *****	

	IFNDEF	LCDport
		ERROR "Declare LCDport in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCDtris
		ERROR "Declare LCDtris in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCD_EN
		ERROR "Declare LCD_EN in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCD_ENtris
		ERROR "Declare LCD_ENtris in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCD_RS
		ERROR "Declare LCD_RS in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCD_RStris
		ERROR "Declare LCD_RStris in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCD_RW
		ERROR "Declare LCD_RW in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCD_RWtris
		ERROR "Declare LCD_RWtris in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCDbusy
		ERROR "#define LCDbusy FLAGreg,0x00 in MAIN PROGRAM"
	ENDIF
	IFNDEF	LCDcflag
		ERROR "#define LCDcflag FLAGreg,0x01 in MAIN PROGRAM"
	ENDIF
	
	; LCD data ports D4 - D7 == 0x00 - 0x03
D7	equ	0x07	; LCD data line 7, for busy flag

;***** REGISTER DECLARATION *****

	IFNDEF	BASE
	  ERROR "Declare BASE (Base address of user file registers) in MAIN PROGRAM"
	ENDIF

LCDbuf	set	BASE+d'2'	; LCD data buffer
LCDtemp	set	BASE+d'3'	; LCD temporary register
LCDtemp2 set	BASE+d'2'	; busy wait loop (re-use LCDbuf)
	
;***** LCD COMMANDS *****

  ;*** Standard LCD COMMANDS for INIT ***	( HI-NIBBLE only )
	; for 4 bit mode: send only one nibble as high-nibble [DB7:DB4]
	CONSTANT  LCDEM8  = b'00110000'	; entry mode set: 8 bit mode, 2 lines
	CONSTANT  LCDEM4  = b'00100000'	; entry mode set: 4 bit mode, 2 lines
	CONSTANT  LCDDZ   = b'10000000'	; set Display Data Ram Address to zero

  ;*** Standard LCD COMMANDS ***		( HI- / LO-NIBBLE )
	; USE THESE COMMANDS BELOW AS FOLLOW: "LCDcmd LCDCLR"
	CONSTANT  LCDCLR  = b'00000001'	; clear display: resets address counter & cursor
	CONSTANT  LCDCH   = b'00000010'	; cursor home
	CONSTANT  LCDCR   = b'00000110'	; entry mode set: cursor moves right, display auto-shift off
	CONSTANT  LCDCL   = b'00000100'	; entry mode set: cursor moves left, display auto-shift off
	CONSTANT  LCDCONT = b'00001100'	; display control: display on, cursor off, blinking off
	CONSTANT  LCDBL = b'00001111'	; display control: display on, cursor oN, blinking oN
	CONSTANT  LCDMCL  = b'00010000'	; cursor/disp control: move cursor left
	CONSTANT  LCDMCR  = b'00010100'	; cursor/disp control: move cursor right
	CONSTANT  LCDSL   = b'00011000'	; cursor/disp control: shift display content left
	CONSTANT  LCDSR   = b'00011100'	; cursor/disp control: shift display content right
	CONSTANT  LCD2L   = b'00101000'	; function set: 4 bit mode, 2 lines, 5x7 dots
	IF (LCDLINENUM == 0x2)
	  CONSTANT  LCDL1 = b'10000000'	; DDRAM address: 0x00, selects line 1 (2xXX LCD)
	  CONSTANT  LCDL2 = b'11000000'	; DDRAM address: 0x40, selects line 2 (2xXX LCD)
	  CONSTANT  LCDL3 = b'10010100'	; (DDRAM address: 0x14, fallback)
	  CONSTANT  LCDL4 = b'11010100'	; (DDRAM address: 0x54, fallback)
	ELSE
	  CONSTANT  LCDL1 = b'10000000'	; DDRAM address: 0x00, selects line 1 (4xXX LCD)
	  CONSTANT  LCDL2 = b'10010100'	; DDRAM address: 0x14, selects line 2 (4xXX LCD)
	  CONSTANT  LCDL3 = b'11000000'	; DDRAM address: 0x40, selects line 3 (4xXX LCD)
	  CONSTANT  LCDL4 = b'11010100'	; DDRAM address: 0x54, selects line 4 (4xXX LCD)
	ENDIF
	; special configuration for EA DIP204-4
	CONSTANT  LCDEXT  = b'00001001'	; extended function set EA DIP204-4
	CONSTANT  LCD2L_A = b'00101100'	; enter ext. function set: 4 bit mode, 2 lines, 5x7 dots
	CONSTANT  LCD2L_B = b'00101000'	; exit ext. function set: 4 bit mode, 2 lines, 5x7 dots
	
;***** MACROS *****

LCDinit	macro
	BANK1
	bcf	LCD_ENtris	; set command lines to output
	bcf	LCD_RStris
	bcf	LCD_RWtris
	movlw	b'00001111'	; data output on low nibble
	andwf	LCDtris,f
	BANK0
	bcf	LCD_EN		; clear LCD clock line
	bcf	LCD_RW		; set write direction
	bcf	LCD_RS		; clear command/data line
	clrLCDport		; reset LCD data lines
	movlw .5	; >= 4 ms @ 4 MHz
	call BITDELAY
	movlw .5	; >= 4 ms @ 4 MHz
	call BITDELAY
	; LCD INITIALIZATION STARTS HERE
	; start in 8 bit mode
	movlw	LCDEM8		; send b'0011' on [DB7:DB4]
	call	LCDxmit		; start in 8 bit mode
	call	LCDclk		; That's while:
	movlw .5
	call BITDELAY	; On POWER UP, the LCD will initialize itself,
				; but after a RESET of the microcontroller without
				; POWER OFF, the 8 bit function mode will reboot
				; the LCD to 4 bit mode safely.

	movlw	LCDDZ		; set DDRAM to zero
	call	LCDxmit
	call	LCDclk
	movlw .5
	call BITDELAY		; ~1 ms @ 4 MHz

	movlw	LCDEM4		; send b'0010' on [DB7:DB4]
	call	LCDxmit		; change to 4 bit mode
	call	LCDclk
	movlw .5
	call BITDELAY		; ~1 ms @ 4 MHz
	
	; now in 4 bit mode, sending two nibbles

	  LCDcmd LCD2L		; function set: 4 bit mode, 2 lines, 5x7 dots
	  LCDcmd LCDCONT	; display control: display on, cursor off, blinking off
	  LCDcmd LCDCLR		; clear display, address counter to zero
	movlw .5 
	call BITDELAY	; wait after LCDCLR until LCD is ready again

	endm

LCDchar	macro	LCDarg		; write ASCII argument to LCD
	movlw	LCDarg
	call	LCDdata
	endm
	
LCDw	macro			; write content of w to LCD
	call	LCDdata
	endm
	
LCDcmd	macro	LCDcommand	; write command to LCD
	movlw	LCDcommand
	call	LCDcomd
	endm
	
LCDline	macro	line_num
	IF (line_num == 1)
	  LCDcmd LCDL1		; first line
	ELSE
	  IF (line_num == 2)
	    LCDcmd LCDL2	; second line
	  ELSE
	    IF (line_num == 3)
	      LCDcmd LCDL3	; third line
	    ELSE
	      IF (line_num == 4)
	        LCDcmd LCDL4	; fourth line
 	      ELSE
	        ERROR "Wrong line number specified in LCDline"
	      ENDIF
	    ENDIF
	  ENDIF
	ENDIF
	endm

LCD_DDAdr macro DDRamAddress
	Local	value = DDRamAddress | b'10000000'	; mask command
	IF (DDRamAddress > 0x67)
	  ERROR "Wrong DD-RAM-Address specified in LCD_DDAdr"
	ELSE
	  movlw	value
	  call	LCDcomd
	ENDIF
	endm

LCD_CGAdr macro	CGRamAddress
	Local	value = CGRamAddress | b'01000000'	; mask command
	IF (CGRamAddress > b'00111111')
	  ERROR "Wrong CG-RAM-Address specified in LCD_CGAdr"
	ELSE
	  movlw	value
	  call	LCDcomd
	ENDIF
	endm

clrLCDport macro		; clear/reset LCD data lines
	movlw	b'00001111'	; get mask
	andwf	LCDport,f	; clear data lines only
	endm

;***** SUBROUTINES *****

	; transmit only lower nibble of w
LCDxmit	movwf	LCDbuf		; store command/data nibble
	; first, clear LCD data lines
	clrLCDport
	; second, move data out to LCD data lines
	movfw	LCDbuf		; get data
	andlw	b'11110000'	; extract only valid part
	iorwf	LCDport,f	; put to LCD data lines
	RETURN

	; transmit command to LCD
LCDcomd bsf	LCDcflag	; set command
	goto	_LCD_wr

	; transmit data to LCD
LCDdata bcf	LCDcflag	; set data
_LCD_wr	movwf	LCDtemp		; store command/data to send
	call	CheckBusy	; check, if ready for next character
	btfss	LCDcflag	; skip if LCD command has been sent
	bsf	LCD_RS		; select data registers
	; send high nibble
	movfw	LCDtemp		; get data
	call	LCDxmit		; transmit nibble
	call	LCDclk
	; send low nibble
	movfw	LCDtemp		; get data
	swapf	LCDtemp,w	; swap high and low nibble, store in w
	call	LCDxmit		; transmit nibble
	call	LCDclk
	; reset LCD controls
	clrLCDport		; reset LCD data lines
	bcf	LCD_RS		; reset command/data register
	bcf	LCD_RW		; reset to write direction
	RETURN

	; clocks LCD data/command
LCDclk	bsf     LCD_EN		; set LCD enable
	; insert LCDSPEED x nops to comply with manufacturer
	; specifications for clock rates above 9 MHz
	VARIABLE CNT_V		; declare intermediate variable
CNT_V = LCDSPEED	; assign pre-defined constant
	WHILE (CNT_V > 0x0)	; perform while loop to insert 'nops'
	  nop			; insert 'nop'
CNT_V -= 1		; decrement
	ENDW
	bcf     LCD_EN
	RETURN

	; busy wait loop, makes busy flag signalling safe
LCDbusyloop
	movlw	LCDBUSYWAIT	; pre-defined constant
	movwf	LCDtemp2
_LCDbusyloop
	decfsz	LCDtemp2,f
	goto	_LCDbusyloop	; busy loop
	RETURN
	
	; routine to poll busy flag from LCD
CheckBusy
	bcf	LCD_RS		; select command registers
	bsf	LCDbusy		; init LCD busy flag
	BANK1
	movlw	b'11110000'
	iorwf	LCDtris,f	; set corresponding ports to inputs
	BANK0
	bsf	LCD_RW		; apply read direction
	nop			; additional safety
_LCDbusy bsf	LCD_EN		; set enable
	; the following busy wait code makes busy flag signalling safe
	call	LCDbusyloop	; (timing relaxation)
	; poll now LCD busy flag
	btfss	LCDport,7	; check busy flag, skip if busy 
	bcf	LCDbusy		; set register flag if not busy
	bcf	LCD_EN
	; the following busy wait code makes busy flag signalling safe
	call	LCDbusyloop	; (timing relaxation)
	; get low nibble, ignore it
	call	LCDclk
	btfsc	LCDbusy		; skip if register flag is not yet cleared
	goto	_LCDbusy
	bcf	LCD_RW		; re-apply write direction
	BANK1
	movlw	b'00001111'
	andwf	LCDtris,f	; set ports to output again
	BANK0
	RETURN


LCDSDDA
		ANDLW	0x07F
		IORLW	0x080		; Function set
		call	LCDcomd
		RETURN


LCDPUTLine MACRO TEXT

	LOCAL MESSAGE
	LOCAL START
	LOCAL EXIT
	LOCAL I=0
	GOTO START
MESSAGE 
		DT TEXT,H'00'
START
	WHILE I<16
		CALL MESSAGE+I
		ADDLW .0
		BZ EXIT
		CALL LCDdata
I=I+1
	ENDW
EXIT
	ENDM