;*** SIGNED 32-BIT INTEGER MATHS ROUTINES FOR PIC16 SERIES DEVELOPED BY ROMUX.COM ***
;
;FUNCTIONS:
;	ADD
;	SUBTRACT
;	MULTIPLY
;	DIVIDE
;	ROUND
;	SQRT
;	BIN2DEC
;	DEC2BIN


;*** 32 BIT SIGNED SUTRACT ***
;REGA - REGB -> REGA
;RETURN CARRY SET IF OVERFLOW

SUBTRACT
	CALL	NEGATEB		;NEGATE REGB
	SKPNC
	RETURN			;OVERFLOW

;*** 32 BIT SIGNED ADD ***
;REGA + REGB -> REGA
;RETURN CARRY SET IF OVERFLOW

ADD	MOVF	REGA3,W		;COMPARE SIGNS
	XORWF	REGB3,W
	MOVWF	MTEMP

	CALL	ADDBA		;ADD REGB TO REGA

	CLRC			;CHECK SIGNS
	MOVF	REGB3,W		;IF SIGNS ARE SAME
	XORWF	REGA3,W		;SO MUST RESULT SIGN
	BTFSS	MTEMP,7		;ELSE OVERFLOW
	ADDLW	0X80
	RETURN

;*** 32 BIT SIGNED MULTIPLY ***
;REGA * REGB -> REGA
;RETURN CARRY SET IF OVERFLOW

MULTIPLY
	CLRF	MTEMP		;RESET SIGN FLAG
	CALL	ABSA		;MAKE REGA POSITIVE
	SKPC
	CALL	ABSB		;MAKE REGB POSITIVE
	SKPNC
	RETURN			;OVERFLOW

	CALL	MOVAC		;MOVE REGA TO REGC
	CALL	CLRA		;CLEAR PRODUCT

	MOVLW	D'31'		;LOOP COUNTER
	MOVWF	MCOUNT

MULOOP	CALL	SLAC		;SHIFT LEFT PRODUCT AND MULTIPLICAND
	
	RLF	REGC3,W		;TEST MSB OF MULTIPLICAND
	SKPNC			;IF MULTIPLICAND BIT IS A 1 THEN
	CALL	ADDBA		;ADD MULTIPLIER TO PRODUCT

	SKPC			;CHECK FOR OVERFLOW
	RLF	REGA3,W
	SKPNC
	RETURN

	DECFSZ	MCOUNT,F	;NEXT
	GOTO	MULOOP

	BTFSC	MTEMP,0		;CHECK RESULT SIGN
	CALL	NEGATEA		;NEGATIVE
	RETURN


;*** 32 BIT SIGNED DIVIDE ***
;REGA / REGB -> REGA
;REMAINDER IN REGC
;RETURN CARRY SET IF OVERFLOW OR DIVISION BY ZERO

DIVIDE	CLRF	MTEMP		;RESET SIGN FLAG
	MOVF	REGB0,W		;TRAP DIVISION BY ZERO
	IORWF	REGB1,W
	IORWF	REGB2,W
	IORWF	REGB3,W
	SUBLW	0
	SKPC
	CALL	ABSA		;MAKE DIVIDEND (REGA) POSITIVE
	SKPC
	CALL	ABSB		;MAKE DIVISOR (REGB) POSITIVE
	SKPNC
	RETURN			;OVERFLOW

	CLRF	REGC0		;CLEAR REMAINDER
	CLRF	REGC1
	CLRF	REGC2
	CLRF	REGC3
	CALL	SLAC		;PURGE SIGN BIT

	MOVLW	D'31'		;LOOP COUNTER
	MOVWF	MCOUNT

DVLOOP	CALL	SLAC		;SHIFT DIVIDEND (REGA) MSB INTO REMAINDER (REGC)

	MOVF	REGB3,W		;TEST IF REMAINDER (REGC) >= DIVISOR (REGB)
	SUBWF	REGC3,W
	SKPZ
	GOTO	DTSTGT
	MOVF	REGB2,W
	SUBWF	REGC2,W
	SKPZ
	GOTO	DTSTGT
	MOVF	REGB1,W
	SUBWF	REGC1,W
	SKPZ
	GOTO	DTSTGT
	MOVF	REGB0,W
	SUBWF	REGC0,W
DTSTGT	SKPC			;CARRY SET IF REMAINDER >= DIVISOR
	GOTO	DREMLT

	MOVF	REGB0,W		;SUBTRACT DIVISOR (REGB) FROM REMAINDER (REGC)
	SUBWF	REGC0,F
	MOVF	REGB1,W
	SKPC
	INCFSZ	REGB1,W
	SUBWF	REGC1,F
	MOVF	REGB2,W
	SKPC
	INCFSZ	REGB2,W
	SUBWF	REGC2,F
	MOVF	REGB3,W
	SKPC
	INCFSZ	REGB3,W
	SUBWF	REGC3,F
	CLRC
	BSF	REGA0,0		;SET QUOTIENT BIT

DREMLT	DECFSZ	MCOUNT,F	;NEXT
	GOTO	DVLOOP

	BTFSC	MTEMP,0		;CHECK RESULT SIGN
	CALL	NEGATEA		;NEGATIVE
	RETURN

;*** ROUND RESULT OF DIVISION TO NEAREST INTEGER ***

ROUND	CLRF	MTEMP		;RESET SIGN FLAG
	CALL	ABSA		;MAKE POSITIVE
	CLRC
	CALL	SLC		;MULTIPLY REMAINDER BY 2
	MOVF	REGB3,W		;TEST IF REMAINDER (REGC) >= DIVISOR (REGB)
	SUBWF	REGC3,W
	SKPZ
	GOTO	RTSTGT
	MOVF	REGB2,W
	SUBWF	REGC2,W
	SKPZ
	GOTO	DTSTGT
	MOVF	REGB1,W
	SUBWF	REGC1,W
	SKPZ
	GOTO	RTSTGT
	MOVF	REGB0,W
	SUBWF	REGC0,W
RTSTGT	SKPC			;CARRY SET IF REMAINDER >= DIVISOR
	GOTO	RREMLT
	INCFSZ	REGA0,F		;ADD 1 TO QUOTIENT
	GOTO	RREMLT
	INCFSZ	REGA1,F
	GOTO	RREMLT
	INCFSZ	REGA2,F
	GOTO	RREMLT
	INCF	REGA3,F
	SKPNZ
	RETURN			;OVERFLOW,RETURN CARRY SET
RREMLT	BTFSC	MTEMP,0		;RESTORE SIGN
	CALL	NEGATEA
	RETURN

;*** 32 BIT SQUARE ROOT ***
;SQRT(REGA) -> REGA
;RETURN CARRY SET IF NEGATIVE

SQRT	RLF	REGA3,W		;TRAP NEGATIVE VALUES
	SKPNC
	RETURN

	CALL	MOVAC		;MOVE REGA TO REGC
	CALL	CLRBA		;CLEAR REMAINDER (REGB) AND ROOT (REGA)

	MOVLW	D'16'		;LOOP COUNTER
	MOVWF	MCOUNT

SQLOOP	RLF	REGC0,F		;SHIFT TWO MSB'S
	RLF	REGC1,F		;INTO REMAINDER
	RLF	REGC2,F
	RLF	REGC3,F
	RLF	REGB0,F
	RLF	REGB1,F
	RLF	REGB2,F
	RLF	REGC0,F
	RLF	REGC1,F
	RLF	REGC2,F
	RLF	REGC3,F
	RLF	REGB0,F
	RLF	REGB1,F
	RLF	REGB2,F

	SETC			;ADD 1 TO ROOT
	RLF	REGA0,F		;ALIGN ROOT
	RLF	REGA1,F
	RLF	REGA2,F

	MOVF	REGA2,W		;TEST IF REMDR (REGB) >= ROOT (REGA)
	SUBWF	REGB2,W
	SKPZ
	GOTO	STSTGT
	MOVF	REGA1,W
	SUBWF	REGB1,W
	SKPZ
	GOTO	STSTGT
	MOVF	REGA0,W
	SUBWF	REGB0,W
STSTGT	SKPC			;CARRY SET IF REMDR >= ROOT
	GOTO	SREMLT

	MOVF	REGA0,W		;SUBTRACT ROOT (REGA) FROM REMDR (REGB)
	SUBWF	REGB0,F
	MOVF	REGA1,W
	SKPC
	INCFSZ	REGA1,W
	SUBWF	REGB1,F
	MOVF	REGA2,W
	SKPC
	INCFSZ	REGA2,W
	SUBWF	REGB2,F
	BSF	REGA0,1		;SET CURRENT ROOT BIT

SREMLT	BCF	REGA0,0		;CLEAR TEST BIT
	DECFSZ	MCOUNT,F	;NEXT
	GOTO	SQLOOP

	CLRC
	RRF	REGA2,F		;ADJUST ROOT ALIGNMENT
	RRF	REGA1,F
	RRF	REGA0,F
	RETURN


;*** 32 BIT SIGNED BINARY TO DECIMAL ***
;REGA -> DIGITS 1 (MSD) TO 10 (LSD) & DSIGN
;DSIGN = 0 IF REGA IS POSITIVE, 1 IF NEGATIVE
;RETURN CARRY SET IF OVERFLOW
;USES FSR REGISTER

BIN2DEC	CLRF	MTEMP		;RESET SIGN FLAG
	CALL	ABSA		;MAKE REGA POSITIVE
	SKPNC
	RETURN			;OVERFLOW

	CALL	CLRDIG		;CLEAR ALL DIGITS

	MOVLW	D'32'		;LOOP COUNTER
	MOVWF	MCOUNT

B2DLOOP	RLF	REGA0,F		;SHIFT MSB INTO CARRY
	RLF	REGA1,F
	RLF	REGA2,F
	RLF	REGA3,F

	MOVLW	DIGIT10
	MOVWF	FSR		;POINTER TO DIGITS
	MOVLW	D'10'		;10 DIGITS TO DO
	MOVWF	DCOUNT

ADJLP	RLF	INDF,F		;SHIFT DIGIT AND CARRY 1 BIT LEFT
	MOVLW	D'10'
	SUBWF	INDF,W		;CHECK AND ADJUST FOR DECIMAL OVERFLOW
	SKPNC
	MOVWF	INDF

	DECF	FSR,F		;NEXT DIGIT
	DECFSZ	DCOUNT,F
	GOTO	ADJLP

	DECFSZ	MCOUNT,F	;NEXT BIT
	GOTO	B2DLOOP

	BTFSC	MTEMP,0		;CHECK SIGN
	BSF	DSIGN,0		;NEGATIVE
	CLRC
	RETURN


;*** 32 BIT SIGNED DECIMAL TO BINARY ***
;DECIMAL DIGIT1 THRO DIGIT(X) & DSIGN -> REGA
;SET DSIGN = 0 FOR POSITIVE, DSIGN = 1 FOR NEGATIVE VALUES
;MOST SIGNIFICANT DIGIT IN DIGIT1
;ENTER THIS ROUTINE WITH DIGIT COUNT IN W REGISTER
;RETURN CARRY SET IF OVERFLOW
;USES FSR REGISTER

DEC2BIN	MOVWF	MTEMP		;SAVE DIGIT COUNT

	MOVLW	D'32'		;OUTER BIT LOOP COUNTER
	MOVWF	MCOUNT

D2BLP1	MOVLW	DIGIT1-1	;SET UP POINTER TO MSD
	MOVWF	FSR
	MOVF	MTEMP,W		;INNER DIGIT LOOP COUNTER
	MOVWF	DCOUNT

	MOVLW	D'10'
	CLRC			;BRING IN '0' BIT INTO MSD

D2BLP2	INCF	FSR,F
	SKPNC
	ADDWF	INDF,F		;ADD 10 IF '1' BIT FROM PREV DIGIT
	RRF	INDF,F		;SHIFT OUT LSB OF DIGIT

	DECFSZ	DCOUNT,F	;NEXT L.S. DIGIT
	GOTO	D2BLP2

	RRF	REGA3,F		;SHIFT IN CARRY FROM DIGITS
	RRF	REGA2,F
	RRF	REGA1,F
	RRF	REGA0,F

	DECFSZ	MCOUNT,F	;NEXT BIT
	GOTO	D2BLP1

	MOVF	INDF,W		;CHECK FOR OVERFLOW
	ADDLW	0XFF
	SKPC
	RLF	REGA3,W
	SKPNC
	RETURN

	BTFSC	DSIGN,0		;CHECK RESULT SIGN
	CALL	NEGATEA		;NEGATIVE
	RETURN


;UTILITY ROUTINES


;ADD REGB TO REGA (UNSIGNED)
;USED BY ADD, MULTIPLY,

ADDBA	MOVF	REGB0,W		;ADD LO BYTE
	ADDWF	REGA0,F

	MOVF	REGB1,W		;ADD MID-LO BYTE
	SKPNC			;NO CARRY_IN, SO JUST ADD
	INCFSZ	REGB1,W		;ADD CARRY_IN TO REGB
	ADDWF	REGA1,F		;ADD AND PROPAGATE CARRY_OUT

	MOVF	REGB2,W		;ADD MID-HI BYTE
	SKPNC
	INCFSZ	REGB2,W
	ADDWF	REGA2,F

	MOVF	REGB3,W		;ADD HI BYTE
	SKPNC
	INCFSZ	REGB3,W
	ADDWF	REGA3,F
	RETURN


;MOVE REGA TO REGC
;USED BY MULTIPLY, SQRT

MOVAC	MOVF	REGA0,W
	MOVWF	REGC0
	MOVF	REGA1,W
	MOVWF	REGC1
	MOVF	REGA2,W
	MOVWF	REGC2
	MOVF	REGA3,W
	MOVWF	REGC3
	RETURN


;CLEAR REGB AND REGA
;USED BY SQRT

CLRBA	CLRF	REGB0
	CLRF	REGB1
	CLRF	REGB2
	CLRF	REGB3

;CLEAR REGA
;USED BY MULTIPLY, SQRT

CLRA	CLRF	REGA0
	CLRF	REGA1
	CLRF	REGA2
	CLRF	REGA3
	RETURN


;CHECK SIGN OF REGA AND CONVERT NEGATIVE TO POSITIVE
;USED BY MULTIPLY, DIVIDE, BIN2DEC, ROUND

ABSA	RLF	REGA3,W
	SKPC
	RETURN			;POSITIVE

;NEGATE REGA
;USED BY ABSA, MULTIPLY, DIVIDE, BIN2DEC, DEC2BIN, ROUND

NEGATEA	MOVF	REGA3,W		;SAVE SIGN IN W
	ANDLW	0X80

	COMF	REGA0,F		;2'S COMPLEMENT
	COMF	REGA1,F
	COMF	REGA2,F
	COMF	REGA3,F
	INCFSZ	REGA0,F
	GOTO	NEGA1
	INCFSZ	REGA1,F
	GOTO	NEGA1
	INCFSZ	REGA2,F
	GOTO	NEGA1
	INCF	REGA3,F
NEGA1
	INCF	MTEMP,F		;FLIP SIGN FLAG
	ADDWF	REGA3,W		;RETURN CARRY SET IF -2147483648
	RETURN


;CHECK SIGN OF REGB AND CONVERT NEGATIVE TO POSITIVE
;USED BY MULTIPLY, DIVIDE

ABSB	RLF	REGB3,W
	SKPC
	RETURN			;POSITIVE

;NEGATE REGB
;USED BY ABSB, SUBTRACT, MULTIPLY, DIVIDE

NEGATEB	MOVF	REGB3,W		;SAVE SIGN IN W
	ANDLW	0X80

	COMF	REGB0,F		;2'S COMPLEMENT
	COMF	REGB1,F
	COMF	REGB2,F
	COMF	REGB3,F
	INCFSZ	REGB0,F
	GOTO	NEGB1
	INCFSZ	REGB1,F
	GOTO	NEGB1
	INCFSZ	REGB2,F
	GOTO	NEGB1
	INCF	REGB3,F
NEGB1
	INCF	MTEMP,F		;FLIP SIGN FLAG
	ADDWF	REGB3,W		;RETURN CARRY SET IF -2147483648
	RETURN


;SHIFT LEFT REGA AND REGC
;USED BY MULTIPLY, DIVIDE, ROUND

SLAC	RLF	REGA0,F
	RLF	REGA1,F
	RLF	REGA2,F
	RLF	REGA3,F
SLC	RLF	REGC0,F
	RLF	REGC1,F
	RLF	REGC2,F
	RLF	REGC3,F
	RETURN


;SET ALL DIGITS TO 0
;USED BY BIN2DEC

CLRDIG	CLRF	DSIGN
	CLRF	DIGIT1
	CLRF	DIGIT2
	CLRF	DIGIT3
	CLRF	DIGIT4
	CLRF	DIGIT5
	CLRF	DIGIT6
	CLRF	DIGIT7
	CLRF	DIGIT8
	CLRF	DIGIT9
	CLRF	DIGIT10
	RETURN

;MOV32REGA

MOV32REG MACRO REG, val
		movlw   val & H'FF'
		movwf   REG
		movlw   val >>D'08' & H'FF'
		movwf   REG+1
		movlw   val >>D'16' & H'FF'
		movwf   REG+2
		movlw   val >>D'24' & H'FF'
		movwf   REG+3
	ENDM	